/* Copyright(C) 2004-2005 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include "snip.h"

#if !defined MAX
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#endif

#if !defined MIN
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

/* These BM/Turbo-BM algolithm is implemented using examples from */
/* EXACT STRING MATCHING ALGORITHMS by Christian Charras - Thierry Lecroq */
/* e-mails: {Christian.Charras, Thierry.Lecroq}@laposte.net */
/* http://www-igm.univ-mlv.fr/~lecroq/string/node14.html */
/* AND Fussy's Homepage */
/* e-mail: fushimi@sd.starcat.ne.jp */
/* http://www2.starcat.ne.jp/~fussy/algo/algo7-4.htm */

#define ASIZE 256               /* character-set size */

typedef struct
{
  int start_offset;
  int end_offset;               /* dup */
  float score;
  sen_snippet_cond *cond;
} sen_bm_search_result;

typedef struct
{
  int start_offset;
  int end_offset;
  int first_keyword_id;
  float score;
} sen_snippet_score;

void
sen_bm_suffixes(const unsigned char *x, int m, int *suff)
{
  int f, g, i;
  f = 0;
  suff[m - 1] = m;
  g = m - 1;
  for (i = m - 2; i >= 0; --i) {
    if (i > g && suff[i + m - 1 - f] < i - g)
      suff[i] = suff[i + m - 1 - f];
    else {
      if (i < g)
        g = i;
      f = i;
      while (g >= 0 && x[g] == x[g + m - 1 - f])
        --g;
      suff[i] = f - g;
    }
  }
}

void
sen_bm_preBmBc(const unsigned char *x, int m, int *bmBc)
{
  int i;
  for (i = 0; i < ASIZE; ++i)
    bmBc[i] = m;
  for (i = 0; i < m - 1; ++i)
    bmBc[(int) x[i]] = m - i - 1;
}

int
sen_bm_preBmGs(const unsigned char *x, int m, int *bmGs)
{
  int i, j, *suff;
  if (!(suff = SEN_MALLOC(sizeof(int) * m))) {
    return 1;
  }
  sen_bm_suffixes(x, m, suff);
  for (i = 0; i < m; ++i)
    bmGs[i] = m;
  j = 0;
  for (i = m - 1; i >= -1; --i)
    if (i == -1 || suff[i] == i + 1)
      for (; j < m - 1 - i; ++j)
        if (bmGs[j] == m)
          bmGs[j] = m - 1 - i;
  for (i = 0; i <= m - 2; ++i)
    bmGs[m - 1 - suff[i]] = m - 1 - i;
  SEN_FREE(suff);
  return 0;
}

int
sen_bm_check_euc(const unsigned char *x, const int y)
{
  const unsigned char *p;
  for (p = x + y - 1; p >= x && *p >= 0x80; p--);
  return ((x + y - p) % 2);
}

int
sen_bm_check_sjis(const unsigned char *x, const int y)
{
  const unsigned char *p;
  for (p = x + y - 1; p >= x; p--)
    if ((*p < 0x81) || (*p > 0x9F && *p < 0xE0) || (*p > 0xEF))
      break;
  return ((x + y - p) % 2);
}

int
sen_bm_BM(sen_snippet_cond * cond, const unsigned char *y, int n,
          sen_encoding encoding, int max_results, sen_bm_search_result * results)
{
  const unsigned char *x;
  int i, j, ok, m, count = 0;
  float score = max_results;
  x = (unsigned char *) cond->keyword->norm;
  m = cond->keyword_len;
  j = 0;
  while (j <= n - m) {
    for (i = m - 1; i >= 0 && x[i] == y[i + j]; --i);
    if (i < 0) {
      switch (encoding) {
      case sen_enc_euc_jp:
        ok = sen_bm_check_euc((unsigned char *) y, j);
        break;
      case sen_enc_sjis:
        ok = sen_bm_check_sjis((unsigned char *) y, j);
        break;
      default:
        ok = 1;
        break;
      }
      if (ok) {
        results[count].start_offset = j;
        results[count].end_offset = j + m;
        results[count].score = score;
        results[count].cond = cond;
        score /= 2;
        if (max_results <= ++count)
          break;
      }
      j += cond->bmGs[0];
    } else
      j += MAX(cond->bmGs[i], cond->bmBc[(int) y[i + j]] - m + 1 + i);
  }
  return count;
}

int
sen_bm_TBM(sen_snippet_cond * cond, const unsigned char *y, int n,
           sen_encoding encoding, int max_results, sen_bm_search_result * results)
{
  const unsigned char *x;
  int bcShift, i, j, shift, u, v, turboShift, ok, m, count = 0;
  float score = max_results;
  x = (unsigned char *) cond->keyword->norm;
  m = cond->keyword_len;
  j = u = 0;
  shift = m;
  while (j <= n - m) {
    i = m - 1;
    while (i >= 0 && x[i] == y[i + j]) {
      --i;
      if (u != 0 && i == m - 1 - shift)
        i -= u;
    }
    if (i < 0) {
      switch (encoding) {
      case sen_enc_euc_jp:
        ok = sen_bm_check_euc((unsigned char *) y, j);
        break;
      case sen_enc_sjis:
        ok = sen_bm_check_sjis((unsigned char *) y, j);
        break;
      default:
        ok = 1;
        break;
      }
      if (ok) {
        results[count].start_offset = j;
        results[count].end_offset = j + m;
        results[count].score = score;
        results[count].cond = cond;
        score /= 2;
        if (max_results <= ++count)
          break;
      }
      shift = cond->bmGs[0];
      u = m - shift;
    } else {
      v = m - 1 - i;
      turboShift = u - v;
      bcShift = cond->bmBc[(int) y[i + j]] - m + 1 + i;
      shift = MAX(turboShift, bcShift);
      shift = MAX(shift, cond->bmGs[i]);
      if (shift == cond->bmGs[i])
        u = MIN(m - shift, v);
      else {
        if (turboShift < bcShift)
          shift = MAX(shift, u + 1);
        u = 0;
      }
    }
    j += shift;
  }
  return count;
}

sen_snippet_cond *
sen_snip_cond_open(const char *keyword, sen_encoding encoding, int flags,
                   const char *opentag, const char *closetag)
{
  sen_snippet_cond *cond;
  if (!keyword || !*keyword) {
    return NULL;
  }
  if (!(cond = SEN_MALLOC(sizeof(sen_snippet_cond)))) {
    return NULL;
  }
  cond->opentag_len = strlen(opentag);
  cond->closetag_len = strlen(closetag);
  if (flags) {
    cond->keyword = sen_nstr_open(keyword, encoding, 0);        /* normalize dup */
  } else {
    sen_nstr *nstr;
    if (!(nstr = SEN_MALLOC(sizeof(sen_nstr)))) {
      sen_snip_cond_close(cond);
      return NULL;
    }
    nstr->norm = SEN_STRDUP(keyword);   /* dup, the label is 'norm' but not normalized  */
    cond->keyword = nstr;
  }
  cond->keyword_len = strlen(cond->keyword->norm);
  cond->opentag = opentag;      /* not copy tags! */
  cond->closetag = closetag;
  if (!(cond->bmGs = SEN_MALLOC(sizeof(int) * cond->keyword_len))) {
    sen_snip_cond_close(cond);
    return NULL;
  }
  if (!(cond->bmBc = SEN_MALLOC(sizeof(int) * ASIZE))) {
    sen_snip_cond_close(cond);
    return NULL;
  }
  if (sen_bm_preBmGs(cond->keyword->norm, cond->keyword_len, cond->bmGs)) {
    sen_snip_cond_close(cond);
    return NULL;
  }
  sen_bm_preBmBc(cond->keyword->norm, cond->keyword_len, cond->bmBc);
  return cond;
}

sen_rc
sen_snip_cond_close(sen_snippet_cond * cond)
{
  if (!cond) {
    return sen_invalid_argument;
  }
  if (cond->keyword) {
    sen_nstr_close(cond->keyword);
  }
  if (cond->bmGs) {
    SEN_FREE(cond->bmGs);
  }
  if (cond->bmBc) {
    SEN_FREE(cond->bmBc);
  }
  SEN_FREE(cond);
  return sen_success;
}

int
offset_compare(sen_bm_search_result * a, sen_bm_search_result * b)
{
  if (a->start_offset > b->start_offset)
    return 1;
  else if (a->start_offset < b->start_offset)
    return -1;
  else {
    if (a->end_offset > b->end_offset)
      return 1;
    else if (a->end_offset < b->end_offset)
      return -1;
    else
      return 0;
  }
}

int
snip_score_compare(sen_snippet_score * a, sen_snippet_score * b)
{
  if (a->score < b->score)
    return 1;
  else if (a->score > b->score)
    return -1;
  else
    return 0;
}

int
sen_snip_find_firstbyte(const char *string, sen_encoding encoding, int offset,
                        int doffset)
{
  switch (encoding) {
  case sen_enc_euc_jp:
    while (!(sen_bm_check_euc((unsigned char *) string, offset)))
      offset += doffset;
    break;
  case sen_enc_sjis:
    if (!(sen_bm_check_sjis((unsigned char *) string, offset)))
      offset += doffset;
    break;
  case sen_enc_utf8:
    while (string[offset] <= (char) 0xc0)
      offset += doffset;
    break;
  }
  return offset;
}

sen_rc
sen_snip_snippet_open(const char *string, sen_encoding encoding, int flags,
                      int width, int cond_len, sen_snippet_cond ** cond,
                      int max_results, int *result_count, char ***results)
{
  int i, j, k, l, string_len, count, total_tagged_len = 0, last_end_offset =
    0, last_last_end_offset;
  char **snippets, *snip;
  sen_bm_search_result *res = NULL;
  sen_snippet_score *score = NULL;
  sen_nstr *nstr = NULL;
  sen_rc ret = sen_success;
  if (!cond_len || !max_results || !results) {
    return sen_invalid_argument;
  }
  *result_count = 0;
  *results = NULL;
  if (!string || !*string) { return sen_invalid_argument; }
  if (flags) {
    nstr = sen_nstr_open(string, encoding, 0);
    string = nstr->norm;
  }

  if (!((res = SEN_MALLOC(sizeof(sen_bm_search_result) * max_results * cond_len))
        && (score = SEN_MALLOC(sizeof(sen_snippet_score) * max_results * cond_len)))) {
    ret = sen_memory_exhausted;
    goto END;
  }
  string_len = strlen(string);  /* byte length, not nstr->length */
  count = 0;

  for (i = 0; i < cond_len; i++)
    count += sen_bm_TBM(cond[i], string, string_len, encoding, max_results, res + count);

  /* offset sort */
  qsort(res, count, sizeof(sen_bm_search_result),
        (int (*)(const void *, const void *)) offset_compare);
  for (i = 0; i < count; i++) {
    if (res[i].start_offset < last_end_offset) {
      score[i].score = -1;
    } else {
      (*result_count)++;
      score[i].score = 0;
      last_last_end_offset = last_end_offset;
      for (j = i; j < count; j++) {
        if (res[j].end_offset > res[i].start_offset + width)
          break;
        /*      if(res[j].end_offset <= res[i].end_offset){ *//* for filter <tag2> in case <tag1>a<tag2>b</tag1>c</tag2> */
        score[i].score += res[j].score;
        last_end_offset = res[j].end_offset;
        total_tagged_len += res[j].cond->opentag_len + res[j].cond->closetag_len;
/*      }*/
      }
      score[i].start_offset =
        MIN((res[i].start_offset + last_end_offset - width) / 2, string_len - width);
      if (score[i].start_offset < last_last_end_offset)
        score[i].start_offset = last_last_end_offset;
      score[i].start_offset =
        sen_snip_find_firstbyte(string, encoding, score[i].start_offset, 1);
      score[i].end_offset = score[i].start_offset + width;
      score[i].end_offset =
        sen_snip_find_firstbyte(string, encoding, score[i].end_offset, -1);
      last_end_offset = score[i].end_offset;

      total_tagged_len += score[i].end_offset - score[i].start_offset + 1;
      score[i].first_keyword_id = i;
    }
  }
  /* score sort */
  qsort(score, count, sizeof(sen_snippet_score),
        (int (*)(const void *, const void *)) snip_score_compare);
  max_results = MIN(*result_count, max_results);
  if (!(snippets = SEN_MALLOC(sizeof(char *) * max_results + total_tagged_len))) {
    ret = sen_memory_exhausted;
    goto END;
  }
  snip = (char *) snippets + sizeof(char *) * max_results;
  for (i = 0; i < max_results; i++) {
    snippets[i] = snip;
    k = score[i].first_keyword_id;
    l = k;
    for (j = score[i].start_offset; j < score[i].end_offset; j++) {
      while (k < count && res[k].start_offset == j) {
        if (res[k].end_offset <= score[i].end_offset) {
          memcpy(snip, res[k].cond->opentag, res[k].cond->opentag_len);
          snip += res[k].cond->opentag_len;
        }
        k++;
      }
      *(snip++) = string[j];
      while (l < count && res[l].end_offset == (j + 1)) {
        memcpy(snip, res[l].cond->closetag, res[l].cond->closetag_len);
        snip += res[l].cond->closetag_len;
        l++;
      }
    }
    *(snip++) = '\0';
  }
  *results = snippets;
END:
  if (flags && nstr) {
    sen_nstr_close(nstr);
  }
  if (res) {
    SEN_FREE(res);
  }
  if (score)
    SEN_FREE(score);
  return sen_success;
}

sen_rc
sen_snip_snippet_close(char **results)
{
  int i;
  if (!results) {
    return sen_invalid_argument;
  }
  SEN_FREE(results);
  return sen_success;
}
